# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
from twisted.internet import task
if(utils.checkTwisted()):
	from twisted.xish.domish import Element
else:
	from tlib.domish import Element
from tlib import msn
import groupchat
import msnw
import config
import debug
import lang




name = "MSN Transport"	# The name of the transport
version = "0.9.3"	# The transport version
mangle = True		# XDB '@' -> '%' mangling
id = "msn"		# The transport identifier



def isGroupJID(jid):
	""" Returns True if the JID passed is a valid groupchat JID (for MSN, does not contain '%') """
	return (jid.find('%') == -1)


	
# This should be set to the name space the registration entries are in, in the xdb spool
namespace = "jabber:iq:register"


def formRegEntry(username, password, nickname):
	""" Returns a domish.Element representation of the data passed. This element will be written to the XDB spool file """
	reginfo = Element((None, "query"))
	reginfo.attributes["xmlns"] = "jabber:iq:register"
	
	userEl = reginfo.addElement("username")
	userEl.addContent(username)
	
	passEl = reginfo.addElement("password")
	passEl.addContent(password)

	nickEl = reginfo.addElement("nick")
	if(nickname): nickEl.addContent(nickname)
	
	return reginfo




def getAttributes(base):
	""" This function should, given a spool domish.Element, pull the username, password,
	and nickname out of it and return them """
	username = ""
	password = ""
	nickname = ""
	for child in base.elements():
		try:
			if(child.name == "username"):
				username = child.__str__()
			elif(child.name == "password"):
				password = child.__str__()
			elif(child.name == "nick"):
				nickname = child.__str__()
		except AttributeError:
			continue
	
	return username, password, nickname





def msn2jid(msnid):
	""" Converts a MSN passport into a JID representation to be used with the transport """
	return msnid.replace('@', '%') + "@" + config.jid

translateAccount = msn2jid # Marks this as the function to be used in jabber:iq:gateway (Service ID Translation)

def jid2msn(jid):
	""" Converts a JID representation of a MSN passport into the original MSN passport """
	return unicode(jid[:jid.find('@')].replace('%', '@'))


def presence2state(show, ptype): 
	""" Converts a Jabber presence into an MSN status code """
	if(ptype == "unavailable"):
		return msn.STATUS_OFFLINE
	elif(show in [None, "online", "chat"]):
		return msn.STATUS_ONLINE
	elif(show in ["dnd"]):
		return msn.STATUS_BUSY
	elif(show in ["away", "xa"]):
		return msn.STATUS_AWAY


def state2presence(state):
	""" Converts a MSN status code into a Jabber presence """
	if(state == msn.STATUS_ONLINE):
		return (None, None)
	elif(state == msn.STATUS_BUSY):
		return ("dnd", None)
	elif(state == msn.STATUS_AWAY):
		return ("away", None)
	elif(state == msn.STATUS_IDLE):
		return ("away", None)
	elif(state == msn.STATUS_BRB):
		return ("away", None)
	elif(state == msn.STATUS_PHONE):
		return ("dnd", None)
	elif(state == msn.STATUS_LUNCH):
		return ("away", None)
	else:
		return (None, "unavailable")




# This class handles groupchats with the legacy protocol
class LegacyGroupchat(groupchat.BaseGroupchat):
	def __init__(self, session, resource, ID=None, existing=False, switchboardSession=None):
		""" Possible entry points for groupchat
			- User starts an empty switchboard session by sending presence to a blank room
			- An existing switchboard session is joined by another MSN user
			- User invited to an existing switchboard session with more than one user
		"""
		groupchat.BaseGroupchat.__init__(self, session, resource, ID)
		if(not existing):
			self.switchboardSession = msnw.GroupchatSwitchboardSession(self, makeSwitchboard=True)
		else:
			self.switchboardSession = switchboardSession
			
		assert(self.switchboardSession != None)
		
		debug.log("LegacyGroupchat: \"%s\" created" % (self.roomJID()))
	
	def removeMe(self):
		self.switchboardSession.removeMe()
		self.switchboardSession = None
		groupchat.BaseGroupchat.removeMe(self)
		debug.log("LegacyGroupchat: \"%s\" destroyed" % (self.roomJID()))
		utils.mutilateMe(self)
	
	def sendLegacyMessage(self, message, noerror):
		debug.log("LegacyGroupchat: \"%s\" sendLegacyMessage(\"%s\")" % (self.roomJID(), message))
		self.switchboardSession.sendMessage(message, noerror)
	
	def sendContactInvite(self, contactJID):
		debug.log("LegacyGroupchat: \"%s\" sendContactInvite(\"%s\")" % (self.roomJID(), contactJID))
		userHandle = jid2msn(contactJID)
		self.switchboardSession.inviteUser(userHandle)



# This class handles most interaction with the legacy protocol
class LegacyConnection(msnw.MSNConnection):
	""" A glue class that connects to the legacy network """
	def __init__(self, username, password, session):
		self.session = session
		self.listSynced = False
		self.initialListVersion = 0

		# Get the latest listVersion to pass to MSNConnection
		result = self.session.pytrans.xdb.request(self.session.jabberID, "msn:listVersion")
		if(result):
			self.initialListVersion = int(str(result))

		# Init the MSN bits
		msnw.MSNConnection.__init__(self, username, password)

		# User typing notification stuff
		self.userTyping = dict() # Indexed by contact MSN ID, stores whether the user is typing to this contact
		# Contact typing notification stuff
		self.contactTyping = dict() # Indexed by contact MSN ID, stores an integer that is incremented at 5 second intervals. If it reaches 3 then the contact has stopped typing. It is set to zero whenever MSN typing notification messages are received
		# Looping function
		self.userTypingSend = task.LoopingCall(self.sendTypingNotifications)
		self.userTypingSend.start(5.0)
		
		import subscription # Is in here to prevent an ImportError loop
		self.subscriptions = subscription.SubscriptionManager(self.session)
	
		debug.log("LegacyConnection: \"%s\" - created" % (self.session.jabberID))
	
	def removeMe(self):
		debug.log("LegacyConnection: \"%s\" - being deleted" % (self.session.jabberID))
	
		self.userTypingSend.stop()
	
		if(self.getContacts()):
			for userHandle in self.getContacts().getContacts():
				msnContact = self.getContacts().getContact(userHandle)
				if(msnContact.status and msnContact.status != msn.STATUS_OFFLINE):
					msnContact.status = msn.STATUS_OFFLINE
					self.sendMSNContactPresence(msnContact)
		
#		msnw.MSNConnection.changeStatus(self, msn.STATUS_OFFLINE, self.session.nickname) # Change our nickname on the MSN network (stops users from appearing offline with a nickname "james - Online")
		
		# Save to XDB the current list version (for fast logins)
		if(self.notificationFactory and self.notificationFactory.contacts):
			listVersion = self.notificationFactory.contacts.version
			el = Element((None, "query"))
			el.addContent(str(listVersion))
			self.session.pytrans.xdb.set(self.session.jabberID, "msn:listVersion", el)
		
		msnw.MSNConnection.removeMe(self)
		self.subscriptions.removeMe()
		self.subscriptions = None
		self.session = None

		utils.mutilateMe(self)
	
	def jidRes(self, resource):
		to = self.session.jabberID
		if(resource):
			to += "/" + resource

		return to
	
	def highestResource(self):
		""" Returns highest priority resource """
		return self.session.highestResource()
	
	def sendMessage(self, dest, resource, body, noerror):
		dest = jid2msn(dest)
		if(self.userTyping.has_key(dest)):
			del self.userTyping[dest]
		msnw.MSNConnection.sendMessage(self, dest, resource, body, noerror)
	
	def buildFriendly(self, status):
		""" Constructs a friendly name from the user's registered nick, and their status message """
		
		if(not config.fancyFriendly):
			if(self.session.nickname and len(self.session.nickname) > 0):
				return self.session.nickname
			else:
				return self.session.jabberID[:self.session.jabberID.find('@')]
		
		if(self.session.nickname and len(self.session.nickname) > 0):
			friendly = self.session.nickname
		else:
			friendly = self.session.jabberID[:self.session.jabberID.find('@')]
		if(status and len(status) > 0):
			friendly += " - "
			friendly += status
		if(len(friendly) > 127):
			friendly = friendly[:124] + "..."
		debug.log("LegacyConnection: buildFriendly(%s) returning \"%s\"" % (self.session.jabberID, friendly))
		return friendly
	
	def msnAlert(self, text, actionurl, subscrurl):
		el = Element((None, "message"))
		el.attributes["to"] = self.session.jabberID
		el.attributes["from"] = config.jid
		el.attributes["type"] = "headline"
		body = el.addElement("body")
		body.addContent(text)
		
		x = el.addElement("x")
		x.attributes["xmlns"] = "jabber:x:oob"
		x.addElement("desc").addContent("More information on this notice.")
		x.addElement("url").addContent(actionurl)

		x = el.addElement("x")
		x.attributes["xmlns"] = "jabber:x:oob"
		x.addElement("desc").addContent("Manage subscriptions to alerts.")
		x.addElement("url").addContent(subscrurl)

		self.session.pytrans.send(el)
	
	def setStatus(self, show, status):
		statusCode = presence2state(show, None)
		msnw.MSNConnection.changeStatus(self, statusCode, self.buildFriendly(status))
	
	def newResourceOnline(self, resource):
		self.sendLists(resource)
	
	def jabberSubscriptionReceived(self, source, subtype):
		self.subscriptions.jabberSubscriptionReceived(source, subtype)
	
	def sendTypingNotifications(self):
		# Send any typing notification messages to the user's contacts
		for contact in self.userTyping.keys():
			if(self.userTyping[contact]):
				self.sendTypingToContact(contact)

		# Send any typing notification messages from contacts to the user
		for contact, resource in self.contactTyping.keys():
			self.contactTyping[(contact, resource)] += 1
			if(self.contactTyping[(contact, resource)] >= 3):
				self.session.sendTypingNotification(self.jidRes(resource), msn2jid(contact), False)
				del self.contactTyping[(contact, resource)]
	
	def gotContactTyping(self, contact, resource):
		# Check if the contact has only just started typing
		if(not self.contactTyping.has_key((contact, resource))):
			self.session.sendTypingNotification(self.jidRes(resource), msn2jid(contact), True)

		# Reset the counter
		self.contactTyping[(contact, resource)] = 0
	
	def userTypingNotification(self, dest, resource, composing):
		dest = jid2msn(dest)
		self.userTyping[dest] = composing
		if(composing): # Make it instant
			self.sendTypingToContact(dest)
	
	
	def sendMSNContactPresence(self, msnContact, to=None):
		if(not to):
			to = self.session.jabberID
		source = msn2jid(msnContact.userHandle)
		show, ptype = state2presence(msnContact.status)
		status = msnContact.screenName.decode("utf-8")
		self.session.sendPresence(to=to, fro=source, show=show, status=status, ptype=ptype)
	
	def sendMSNUserPresence(self, userHandle, to=None):
		msnContact = self.getContacts().getContact(userHandle)
		if(msnContact):
			self.sendMSNContactPresence(msnContact, to)
	
	def sendLists(self, resource):
		""" Sends a copy of the MSN contact presences to this resource """
		debug.log("LegacyConnection: \"%s\" - sendLists(\"%s\")" % (self.session.jabberID, resource))
		fulljid = self.session.jabberID
		if(resource):
			fulljid += "/" + resource
		if(self.getContacts()):
			for userHandle in self.getContacts().getContacts():
				self.sendMSNUserPresence(userHandle, fulljid)
	
	def listSynchronized(self):
		if(self.session):
			self.session.sendPresence(to=self.session.jabberID, fro=config.jid)
			self.subscriptions.syncJabberLegacyLists()
			self.listSynced = True
			self.subscriptions.flushSubscriptionBuffer()
	
	def gotMessage(self, remoteUser, resource, text):
		source = msn2jid(remoteUser)
		text = text.decode("utf-8")
		self.session.sendMessage(self.jidRes(resource), fro=source, body=text, mtype="chat")
	
	def loggedIn(self):
		if(self.session):
			debug.log("LegacyConnection: \"%s\" - loggedIn()" % (self.session.jabberID))
			self.session.ready = True
	
	def contactStatusChanged(self, remoteUser):
		if(self.session): # Make sure the transport isn't shutting down
			debug.log("LegacyConnection: \"%s\" - contactStatusChanged(\"%s\")" % (self.session.jabberID, remoteUser))
			self.sendMSNUserPresence(remoteUser)
	
	def ourStatusChanged(self, statusCode):
		# Send out a new presence packet to the Jabber user so that the MSN-t icon changes
		if(self.session):
			source = config.jid
			to = self.session.jabberID
			show, ptype = state2presence(statusCode)
			debug.log("LegacyConnection: \"%s\" - ourStatusChanged(\"%s\")" % (self.session.jabberID, statusCode))
			self.session.sendPresence(to=to, fro=source, show=show)
	
	def userMapping(self, passport, jid):
		text = lang.get(self.session.lang).userMapping % (passport, jid)
		self.session.sendMessage(to=self.session.jabberID, fro=msn2jid(passport), body=text)
	
	def userAddedMe(self, userHandle):
		self.subscriptions.msnContactAddedMe(userHandle)
	
	def userRemovedMe(self, userHandle):
		self.subscriptions.msnContactRemovedMe(userHandle)
	
	def serverGoingDown(self):
		if(self.session):
			self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=lang.get(self.session.lang).msnMaintenance)
	
	def multipleLogin(self):
		if(self.session):
			self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=lang.get(self.session.lang).msnMultipleLogin)
			self.session.removeMe()
	
	def accountNotVerified(self):
		if(self.session):
			text = lang.get(self.session.lang).msnNotVerified % (self.session.username)
			self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text)
	
	def loginFailure(self, message):
		if(self.session):
			text = lang.get(self.session.lang).msnLoginFailure % (self.session.username)
			self.session.sendErrorMessage(to=self.session.jabberID, fro=config.jid, etype="auth", condition="not-authorized", explanation=text, body="Login Failure")
	
	def failedMessage(self, remoteUser, message):
		if(self.session):
			fro = msn2jid(remoteUser)
			self.session.sendErrorMessage(to=self.session.jabberID, fro=fro, etype="wait", condition="recipient-unavailable", explanation=lang.get(self.session.lang).msnFailedMessage, body=message)
	
	def initialEmailNotification(self, inboxunread, foldersunread):
		text = lang.get(self.session.lang).msnInitialMail % (inboxunread, foldersunread)
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text, mtype="headline")
	
	def realtimeEmailNotification(self, mailfrom, fromaddr, subject):
		text = lang.get(self.session.lang).msnRealtimeMail % (mailfrom, fromaddr, subject)
		self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text, mtype="headline")
		
	def connectionLost(self, reason):
		if(self.session):
			debug.log("LegacyConnection: \"%s\" - connectionLost(\"%s\")" % (self.session.jabberID, reason))
			text = lang.get(self.session.lang).msnDisconnected % ("Error") # FIXME, a better error would be nice =P
			self.session.sendMessage(to=self.session.jabberID, fro=config.jid, body=text)
			self.session.removeMe() # Tear down the session


