#!/usr/local/bin/python
# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
from twisted.internet import reactor, task
from twisted.internet.defer import Deferred
import twisted.python.log
if(utils.checkTwisted()):
	from twisted.words.protocols.jabber import component, jid
	from twisted.xish.domish import Element
else:
	from tlib.jabber import component, jid
	from tlib.domish import Element

import os
if(os.name == "posix"):
	import signal
import sys
reload(sys)
sys.setdefaultencoding("utf-8")
import types

# Must load config before everything else
import config
import xmlconfig
xmlconfig.reloadConfig()

import xdb
import session
import jabw
import disco
import register
import misciq
import lang
import debug
import legacy

#import gc
#gc.set_debug(gc.DEBUG_COLLECTABLE | gc.DEBUG_UNCOLLECTABLE | gc.DEBUG_INSTANCES | gc.DEBUG_OBJECTS)


class PyTransport(component.Service):
	def __init__(self):
		debug.log("PyTransport: Service starting up")
		
		# Discovery, as well as some builtin features
		self.discovery = disco.ServerDiscovery(self)
		self.discovery.addIdentity("gateway", legacy.id, legacy.name)
		self.discovery.addIdentity("conference", "text", legacy.name + " Chatrooms")
		self.discovery.addFeature("http://jabber.org/protocol/muc", None) # So that clients know you can create groupchat rooms on the server
		
		self.xdb = xdb.XDB(config.jid, legacy.mangle)
		if(config.allowRegister):
			self.registermanager = register.RegisterManager(self)
		self.gatewayTranslator = misciq.GatewayTranslator(self)
		self.versionTeller = misciq.VersionTeller(self)

		self.xmlstream = None
		self.sessions = {}
		
		# Groupchat ID handling
		self.lastID = 0
		self.reservedIDs = []

		# Message IDs
		self.messageID = 0
		
		self.loopCheckSessions = task.LoopingCall(self.loopCheckSessionsCall)
		self.loopCheckSessions.start(60.0) # call every ten seconds
		
		# Display active sessions if debug mode is on
		if(config.debugOn):
			self.loop = task.LoopingCall(self.loopCall)
			self.loop.start(60.0) # call every 60 seconds
			twisted.python.log.addObserver(self.exceptionLogger)
		
	
	def removeMe(self):
		debug.log("PyTransport: Service shutting down")
		dic = utils.copyDict(self.sessions)
		for session in dic:
			dic[session].removeMe()
	
	def exceptionLogger(self, *kwargs):
		if(len(config.debugLog) > 0):
			kwargs = kwargs[0]
			if(kwargs.has_key("failure")):
				failure = kwargs["failure"]
				failure.printTraceback(debug) # Pass debug as a pretend file object because it implements the write method
				if(config.debugLog):
					debug.flushDebugSmart()
					print "Exception occured! Check the log!"
	
	def makeMessageID(self):
		self.messageID += 1
		return str(self.messageID)
	
	def makeID(self):
		newID = "r" + str(self.lastID)
		self.lastID += 1
		if(self.reservedIDs.count(newID) > 0):
			# Ack, it's already used.. Try again
			return self.makeID()
		else:
			return newID
	
	def reserveID(self, ID):
		self.reservedIDs.append(ID)
	
	def loopCall(self):
		if(len(self.sessions) > 0):
			debug.log("Sessions:")
			for key in self.sessions:
				debug.log("\t" + self.sessions[key].jabberID)
	
	def loopCheckSessionsCall(self):
		if(len(self.sessions) > 0):
			oldDict = utils.copyDict(self.sessions)
			self.sessions = {}
			for key in oldDict:
				session = oldDict[key]
				if(not session.alive):
					debug.log("Ghost session %s found. This shouldn't happen. Trace" % (session.jabberID))
					# Don't add it to the new dictionary. Effectively removing it
				else:
					self.sessions[key] = session
	
	def componentConnected(self, xmlstream):
		debug.log("PyTransport: Connected to main Jabberd server")
		self.xmlstream = xmlstream
		self.xmlstream.addObserver("/iq", self.discovery.onIq)
		self.xmlstream.addObserver("/presence", self.onPresence)
		self.xmlstream.addObserver("/message", self.onMessage)
		self.xmlstream.addObserver("/route", self.onRouteMessage)
	
	def componentDisconnected(self):
		debug.log("PyTransport: Disconnected from main Jabberd server")
		self.xmlstream = None
	
	def onRouteMessage(self, el):
		for child in el.elements():
			if(child.name == "message"):
				self.onMessage(child)
			elif(child.name == "presence"):
				self.onPresence(child)
			elif(child.name == "iq"):
				self.discovery.onIq(child)
	
	def onMessage(self, el):
		fro = el.getAttribute("from")
		froj = jid.JID(fro.lower())
		to = el.getAttribute("to")
#		if(to.find('@') < 0): return
		mtype = el.getAttribute("type")
		ulang = utils.getLang(el)
		body = None
		for child in el.elements():
			if(child.name == "body"):
				body = child.__str__()
		if(self.sessions.has_key(froj.userhost())):
			self.sessions[froj.userhost()].onMessage(el)
		elif(mtype != "error"):
			debug.log("PyTrans: Sending error response to a message outside of session.")
			jabw.sendErrorMessage(self, fro, to, "auth", "not-authorized", lang.get(ulang).notLoggedIn, body)
	
	def onPresence(self, el):
		fro = el.getAttribute("from")
		ptype = el.getAttribute("type")
		froj = jid.JID(fro.lower())
		to = el.getAttribute("to")
		toj = jid.JID(to.lower())
		ulang = utils.getLang(el)
		if(self.sessions.has_key(froj.userhost())):
			self.sessions[froj.userhost()].onPresence(el)
		else:
			if(to.find('@') < 0):
				# If the presence packet is to the transport (not a user) and there isn't already a session
				if(el.getAttribute("type") in [None, ""]): # Don't create a session unless they're sending available presence
					debug.log("PyTransport: Attempting to create a new session \"%s\"" % (froj.userhost()))
					s = session.makeSession(self, froj.userhost(), ulang)
					if(s):
						self.sessions[froj.userhost()] = s
						debug.log("PyTransport: New session created \"%s\"" % (froj.userhost()))
						# Send the first presence
						s.onPresence(el)
					else:
						debug.log("PyTransport: Failed to create session \"%s\"" % (froj.userhost()))
						jabw.sendMessage(self, to=froj.userhost(), fro=config.jid, body=lang.get(ulang).notRegistered)
				
				elif(el.getAttribute("type") != "error"):
					debug.log("PyTransport: Sending unavailable presence to non-logged in user \"%s\"" % (froj.userhost()))
					pres = Element((None, "presence"))
					pres.attributes["from"] = to
					pres.attributes["to"] = fro
					pres.attributes["type"] = "unavailable"
					self.send(pres)
					return
			
			elif(ptype in ["subscribe", "subscribed", "unsubscribe", "unsubscribed"]):
				# They haven't logged in, and are trying to change subscription to a user
				# Lets log them in and then do it
				debug.log("PyTransport: Attempting to create a session to do subscription stuff %s" % (froj.userhost()))
				s = session.makeSession(self, froj.userhost(), ulang)
				if(s):
					self.sessions[froj.userhost()] = s
					debug.log("PyTransport: New session created \"%s\"" % (froj.userhost()))
					# Tell the session there's a new resource
					s.handleResourcePresence(froj.userhost(), froj.resource, toj.userhost(), toj.resource, 0, None, None, None)
					# Send this subscription
					s.onPresence(el)


class App:
	def __init__(self):
		# Check that there isn't already a PID file
		if(os.path.isfile(utils.doPath(config.pid))):
			pf = open(utils.doPath(config.pid))
			pid = int(str(pf.readline().strip()))
			pf.close()
			if(os.name == "posix"):
				try:
					os.kill(pid, signal.SIGHUP)
					self.alreadyRunning()
				except OSError:
					# The process is still up
					pass
			else:
				self.alreadyRunning()
		
		# Create a PID file
		pid = str(os.getpid())
		pf = file(utils.doPath(config.pid), 'w')
		pf.write("%s\n" % pid);
		pf.close()
	
		self.c = component.buildServiceManager(config.jid, config.secret, "tcp:%s:%s" % (config.mainServer, config.port))
		self.transportSvc = PyTransport()
		self.transportSvc.setServiceParent(self.c)
		self.c.startService()
		reactor.addSystemEventTrigger('before', 'shutdown', self.shuttingDown)
	
	def alreadyRunning(self):
		print "There is already a transport instance running with this configuration."
		print "Exiting..."
		sys.exit(1)
	
	def shuttingDown(self):
		self.transportSvc.removeMe()
		def cb(ignored=None):
			os.remove(utils.doPath(config.pid))
		d = Deferred()
		d.addCallback(cb)
		reactor.callLater(3.0, d.callback, None)
		return d



def SIGHUPstuff(*args):
	xmlconfig.reloadConfig()
	debug.reopenFile()


if(__name__ == "__main__"):
	# Set SIGHUP to reload the config file & close & open debug file
	if(os.name == "posix"):
		signal.signal(signal.SIGHUP, SIGHUPstuff)

	app = App()
	reactor.run()


